import copy
from pathlib import Path

import torch
import torch.nn as nn
from torchvision import models as torch_models

from tonks.vision.helpers import _dense_block, _Identity


class ResnetForMultiTaskClassification(nn.Module):
    """
    PyTorch image attribute model. This model allows you to load
    in some pretrained tasks in addition to creating new ones.

    Examples
    --------
    To instantiate a completely new instance of ResnetForMultiTaskClassification
    and load the weights into this architecture you can set `pretrained` to True::

        model = ResnetForMultiTaskClassification(
            new_task_dict=new_task_dict,
            load_pretrained_resnet = True
        )

        # DO SOME TRAINING

        model.save(SOME_FOLDER, SOME_MODEL_ID)

    To instantiate an instance of ResnetForMultiTaskClassification that has layers for
    pretrained tasks and new tasks, you would do the following::

        model = ResnetForMultiTaskClassification(
            pretrained_task_dict=pretrained_task_dict,
            new_task_dict=new_task_dict
        )

        model.load(SOME_FOLDER, SOME_MODEL_DICT)

        # DO SOME TRAINING

    Parameters
    ----------
    pretrained_task_dict: dict
        dictionary mapping each pretrained task to the number of labels it has
    new_task_dict: dict
        dictionary mapping each new task to the number of labels it has
    load_pretrained_resnet: boolean
        flag for whether or not to load in pretrained weights for ResNet50.
        useful for the first round of training before there are fine tuned weights
    """

    def __init__(self, pretrained_task_dict=None, new_task_dict=None, load_pretrained_resnet=False):
        super(ResnetForMultiTaskClassification, self).__init__()

        self.resnet = torch_models.resnet50(pretrained=load_pretrained_resnet)
        self.resnet.fc = _Identity()

        self.dense_layers = nn.Sequential(
            _dense_block(2048*2, 1024, 2e-3),
            _dense_block(1024, 512, 2e-3),
            _dense_block(512, 256, 2e-3),
        )

        if pretrained_task_dict is not None:
            pretrained_layers = {}
            for key, task_size in pretrained_task_dict.items():
                pretrained_layers[key] = nn.Linear(256, task_size)
            self.pretrained_classifiers = nn.ModuleDict(pretrained_layers)
        if new_task_dict is not None:
            new_layers = {}
            for key, task_size in new_task_dict.items():
                new_layers[key] = nn.Linear(256, task_size)
            self.new_classifiers = nn.ModuleDict(new_layers)

    def forward(self, x):
        """
        Defines forward pass for image model

        Parameters
        ----------
        x: dict of image tensors containing tensors for
        full and cropped images. the full image tensor
        has the key 'full_img' and the cropped tensor has
        the key 'crop_img'

        Returns
        ----------
        A dictionary mapping each task to its logits
        """
        full_img = self.resnet(x['full_img']).squeeze()
        crop_img = self.resnet(x['crop_img']).squeeze()

        full_crop_combined = torch.cat((full_img, crop_img), 1)

        dense_layer_output = self.dense_layers(full_crop_combined)

        logit_dict = {}
        if hasattr(self, 'pretrained_classifiers'):
            for key, classifier in self.pretrained_classifiers.items():
                logit_dict[key] = classifier(dense_layer_output)
        if hasattr(self, 'new_classifiers'):
            for key, classifier in self.new_classifiers.items():
                logit_dict[key] = classifier(dense_layer_output)

        return logit_dict

    def freeze_core(self):
        """Freeze all core model layers"""
        for param in self.resnet.parameters():
            param.requires_grad = False

    def freeze_dense(self):
        """Freeze all core model layers"""
        for param in self.dense_layers.parameters():
            param.requires_grad = False

    def freeze_all_pretrained(self):
        """Freeze pretrained classifier layers and core model layers"""
        self.freeze_core()
        self.freeze_dense()
        if hasattr(self, 'pretrained_classifiers'):
            for param in self.pretrained_classifiers.parameters():
                param.requires_grad = False
        else:
            print('There are no pretrained_classifier layers to be frozen.')

    def unfreeze_pretrained_classifiers(self):
        """Unfreeze pretrained classifier layers"""
        if hasattr(self, 'pretrained_classifiers'):
            for param in self.pretrained_classifiers.parameters():
                param.requires_grad = True
        else:
            print('There are no pretrained_classifier layers to be unfrozen.')

    def unfreeze_pretrained_classifiers_and_core(self):
        """Unfreeze pretrained classifiers and core model layers"""
        for param in self.resnet.parameters():
            param.requires_grad = True
        for param in self.dense_layers.parameters():
            param.requires_grad = True
        self.unfreeze_pretrained_classifiers()

    def save(self, folder, model_id):
        """
        Saves the model state dicts to a specific folder.
        Each part of the model is saved separately to allow for
        new classifiers to be added later.

        Note: if the model has `pretrained_classifiers` and `new_classifers`,
        they will be combined into the `pretrained_classifiers_dict`.

        Parameters
        ----------
        folder: str or Path
            place to store state dictionaries
        model_id: int
            unique id for this model

        Side Effects
        ------------
        saves three files:
            - folder / f'resnet_dict_{model_id}.pth'
            - folder / f'dense_layers_dict_{model_id}.pth'
            - folder / f'pretrained_classifiers_dict_{model_id}.pth'
        """
        if not hasattr(self, 'pretrained_classifiers'):
            classifiers_to_save = copy.deepcopy(self.new_classifiers)
        else:
            # PyTorch's update method isn't working because it doesn't think ModuleDict is a Mapping
            classifiers_to_save = copy.deepcopy(self.pretrained_classifiers)
            if hasattr(self, 'new_classifiers'):
                for key, module in self.new_classifiers.items():
                    classifiers_to_save[key] = module

        folder = Path(folder)
        folder.mkdir(parents=True, exist_ok=True)

        torch.save(
            self.resnet.state_dict(),
            folder / f'resnet_dict_{model_id}.pth'
        )
        torch.save(
            self.dense_layers.state_dict(),
            folder / f'dense_layers_dict_{model_id}.pth'
        )

        torch.save(
            classifiers_to_save.state_dict(),
            folder / f'pretrained_classifiers_dict_{model_id}.pth'
        )

    def load(self, folder, model_id):
        """
        Loads the model state dicts from a specific folder.

        Parameters
        ----------
        folder: str or Path
            place where state dictionaries are stored
        model_id: int
            unique id for this model

        Side Effects
        ------------
        loads from three files:
            - folder / f'resnet_dict_{model_id}.pth'
            - folder / f'dense_layers_dict_{model_id}.pth'
            - folder / f'pretrained_classifiers_dict_{model_id}.pth'
        """
        folder = Path(folder)

        if torch.cuda.is_available():
            self.resnet.load_state_dict(torch.load(folder / f'resnet_dict_{model_id}.pth'))
            self.dense_layers.load_state_dict(
                torch.load(folder / f'dense_layers_dict_{model_id}.pth'))
            self.pretrained_classifiers.load_state_dict(
                torch.load(folder / f'pretrained_classifiers_dict_{model_id}.pth')
            )
        else:
            self.resnet.load_state_dict(
                torch.load(
                    folder / f'resnet_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )
            self.dense_layers.load_state_dict(
                torch.load(
                    folder / f'dense_layers_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )
            self.pretrained_classifiers.load_state_dict(
                torch.load(
                    folder / f'pretrained_classifiers_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )

    def export(self, folder, model_id, model_name=None):
        """
        Exports the entire model state dict to a specific folder.
        Note: if the model has `pretrained_classifiers` and `new_classifiers`,
        they will be combined into the `pretrained_classifiers` attribute before being saved.

        Parameters
        ----------
        folder: str or Path
            place to store state dictionaries
        model_id: int
            unique id for this model
        model_name: str (defaults to None)
            Name to store model under, if None, will default to `multi_task_bert_{model_id}.pth`

        Side Effects
        ------------
        saves one file:
            - folder / model_name
        """
        if hasattr(self, 'new_classifiers'):
            hold_new_classifiers = copy.deepcopy(self.new_classifiers)
        else:
            hold_new_classifiers = None

        hold_pretrained_classifiers = None
        if not hasattr(self, 'pretrained_classifiers'):
            self.pretrained_classifiers = copy.deepcopy(self.new_classifiers)
        else:
            hold_pretrained_classifiers = copy.deepcopy(self.pretrained_classifiers)
            # PyTorch's update method isn't working because it doesn't think ModuleDict is a Mapping
            if hasattr(self, 'new_classifiers'):
                for key, module in self.new_classifiers.items():
                    self.pretrained_classifiers[key] = module
        if hasattr(self, 'new_classifiers'):
            del self.new_classifiers

        if model_name is None:
            model_name = f'multi_task_resnet_{model_id}.pth'

        folder = Path(folder)
        folder.mkdir(parents=True, exist_ok=True)

        torch.save(
            self.state_dict(),
            folder / model_name
        )
        if hold_pretrained_classifiers is not None:
            self.pretrained_classifiers = hold_pretrained_classifiers
        else:
            del self.pretrained_classifiers

        if hold_new_classifiers is not None:
            self.new_classifiers = hold_new_classifiers



import random


class MultiDatasetLoader(object):
    """
    Load datasets for multiple tasks

    Parameters
    ----------
    loader_dict: dict
        dictonary of DataLoaders
    shuffle: Boolean (defaults to True)
        Flag for whether or not to shuffle the data
    """
    def __init__(self, loader_dict, shuffle=True):
        self.loader_dict = loader_dict
        self.shuffle = shuffle

        total_samples = 0
        for key in self.loader_dict.keys():
            total_samples += len(self.loader_dict[key].dataset)

        self.total_samples = total_samples

    def __iter__(self):
        named_batches = []
        iterators = {}

        for key in self.loader_dict.keys():
            current_batches = [key] * len(self.loader_dict[key])
            named_batches += current_batches
            iterators[key] = iter(self.loader_dict[key])

        if self.shuffle:
            random.shuffle(named_batches)

        for key in named_batches:
            yield key, next(iterators[key])

    def __len__(self):
        num_batches = 0

        for key in self.loader_dict.keys():
            num_batches += len(self.loader_dict[key])

        return num_batches



from torchvision import transforms

imagenet_rgb_means = [0.485, 0.456, 0.406]
imagenet_rgb_std = [0.229, 0.224, 0.225]
resnet_img_size = 224

full_img_transforms = {
    'train': transforms.Compose([
        transforms.Resize((256, 256)),
        transforms.RandomResizedCrop(resnet_img_size),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
        transforms.Normalize(imagenet_rgb_means, imagenet_rgb_std)
    ]),
    'val': transforms.Compose([
        transforms.Resize((resnet_img_size, resnet_img_size)),
        transforms.ToTensor(),
        transforms.Normalize(imagenet_rgb_means, imagenet_rgb_std)
    ]),
}

cropped_transforms = {
    'train': transforms.Compose([
        transforms.Resize((256, 256)),
        transforms.RandomHorizontalFlip(),
        transforms.RandomVerticalFlip(),
        transforms.RandomRotation(30),
        transforms.RandomResizedCrop(resnet_img_size),
        transforms.ToTensor(),
        transforms.Normalize(imagenet_rgb_means, imagenet_rgb_std)
    ]),
    'val': transforms.Compose([
        transforms.Resize((resnet_img_size, resnet_img_size)),
        transforms.ToTensor(),
        transforms.Normalize(imagenet_rgb_means, imagenet_rgb_std)
    ]),
}



import copy
import time

from fastprogress.fastprogress import format_time, master_bar, progress_bar
import numpy as np
import torch
import torch.nn.functional as F


class MultiTaskLearner(object):
    """
    Class to encapsulate training and validation steps for a pipeline. Based off the fastai learner.

    Parameters
    ----------
    model: torch.nn.Module
        PyTorch model to use with the Learner
    train_dataloader: MultiDatasetLoader
        dataloader for all of the training data
    val_dataloader: MultiDatasetLoader
        dataloader for all of the validation data
    task_dict: dict
        dictionary with all of the tasks as keys and the number of unique labels as the values
    """
    def __init__(self, model, train_dataloader, val_dataloader, task_dict):
        self.model = model
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.task_dict = task_dict
        self.tasks = [*task_dict]

    def fit(
        self,
        num_epochs,
        scheduler,
        step_scheduler_on_batch,
        loss_function,
        optimizer,
        device='cuda:0',
        best_model=False
    ):
        """
        Fit the PyTorch model

        Parameters
        ----------
        num_epochs: int
            number of epochs to train
        scheduler: torch.optim.lr_scheduler
            PyTorch learning rate scheduler
        step_scheduler_on_batch: bool
            flag of whether to step scheduler on batch (if True) or on epoch (if False)
        loss_function: function
            function to calculate loss with in model
        optimizer: torch.optim
            PyTorch optimzer
        device: str (defaults to 'cuda:0')
            device to run calculations on
        best_model: bool (defaults to `False`)
            flag to save best model from a single `fit` training run based on validation loss
            The default is `False`, which will keep the final model from the training run.
            `True` will keep the best model from the training run instead of the model
            from the final epoch of the training cycle.
        """
        self.model = self.model.to(device)

        current_best_loss = np.iinfo(np.intp).max

        pbar = master_bar(range(num_epochs))
        headers = ['train_loss', 'val_loss']
        for task in self.tasks:
            headers.append(f'{task}_train_loss')
            headers.append(f'{task}_val_loss')
            headers.append(f'{task}_acc')
        headers.append('time')
        pbar.write(headers, table=True)

        for epoch in pbar:
            start_time = time.time()
            self.model.train()

            training_loss_dict = {task: 0.0 for task in self.tasks}

            overall_training_loss = 0.0

            for step, batch in enumerate(progress_bar(self.train_dataloader, parent=pbar)):
                task_type, (x, y) = batch
                x = self._return_input_on_device(x, device)
                y = y.type(torch.LongTensor)
                y = y.to(device)
                num_rows = self._get_num_rows(x)

                output = self.model(x)

                current_loss = loss_function(output[task_type], y)

                scaled_loss = current_loss.item() * num_rows

                training_loss_dict[task_type] += scaled_loss

                overall_training_loss += scaled_loss

                optimizer.zero_grad()
                current_loss.backward()
                optimizer.step()
                if step_scheduler_on_batch:
                    scheduler.step()

            if not step_scheduler_on_batch:
                scheduler.step()

            overall_training_loss = overall_training_loss/self.train_dataloader.total_samples

            for task in self.tasks:
                training_loss_dict[task] = (
                    training_loss_dict[task]
                    / len(self.train_dataloader.loader_dict[task].dataset)
                )

            overall_val_loss, val_loss_dict, accuracies = self.validate(
                loss_function,
                device,
                pbar
            )

            str_stats = []
            stats = [overall_training_loss, overall_val_loss]
            for stat in stats:
                str_stats.append(
                    'NA' if stat is None else str(stat) if isinstance(stat, int) else f'{stat:.6f}'
                )

            for task in self.tasks:
                str_stats.append(f'{training_loss_dict[task]:.6f}')
                str_stats.append(f'{val_loss_dict[task]:.6f}')
                str_stats.append(f"{accuracies[task]['accuracy']:.6f}")

            str_stats.append(format_time(time.time() - start_time))

            pbar.write(str_stats, table=True)

            if best_model and overall_val_loss < current_best_loss:
                current_best_loss = overall_val_loss
                best_model_wts = copy.deepcopy(self.model.state_dict())
                best_model_epoch = epoch

        if best_model:
            self.model.load_state_dict(best_model_wts)
            print(f'Epoch {best_model_epoch} best model saved with loss of {current_best_loss}')

    def validate(self, loss_function, device='cuda:0', pbar=None):
        """
        Evaluate the model on a validation set

        Parameters
        ----------
        loss_function: function
            function to calculate loss with in model
        device: str (defaults to 'cuda:0')
            device to run calculations on
        pbar: fast_progress progress bar (defaults to None)
            parent progress bar for all epochs

        Returns
        -------
        overall_val_loss: float
            overall validation loss for all tasks
        val_loss_dict: dict
            dictionary of validation losses for individual tasks
        accuracies: dict
            accuracy measures for individual tasks
        """
        val_loss_dict = {task: 0.0 for task in self.tasks}
        accuracies = {task: {'num_items': 0.0, 'num_correct': 0.0} for task in self.tasks}

        overall_val_loss = 0.0

        self.model.eval()

        with torch.no_grad():
            for step, batch in enumerate(
                progress_bar(self.val_dataloader, parent=pbar, leave=(pbar is not None))
            ):
                task_type, (x, y) = batch
                x = self._return_input_on_device(x, device)
                y = y.type(torch.LongTensor)
                y = y.to(device)
                num_rows = self._get_num_rows(x)

                output = self.model(x)

                current_loss = loss_function(
                    output[task_type], y
                ).item() * num_rows

                val_loss_dict[task_type] += current_loss

                overall_val_loss += current_loss

                accuracies[task_type]['num_correct'] += torch.sum(
                    torch.max(output[task_type], 1)[1] == y
                ).item()

                accuracies[task_type]['num_items'] += num_rows

        overall_val_loss = overall_val_loss/self.val_dataloader.total_samples

        for task in self.tasks:
            val_loss_dict[task] = (
                val_loss_dict[task]
                / len(self.val_dataloader.loader_dict[task].dataset)
            )

        for task in accuracies.keys():
            accuracies[task]['accuracy'] = (
                accuracies[task]['num_correct'] / accuracies[task]['num_items']
            )

        return overall_val_loss, val_loss_dict, accuracies

    def get_val_preds(self, device='cuda:0'):
        """
        Return true labels and predictions for data in self.val_dataloaders

        Parameters
        ----------
        device: str (defaults to 'cuda:0')
            device to run calculations on

        Returns
        -------
        Dictionary with dictionary for each task type:
            'y_true': numpy array of true labels, shape: (num_rows,)
            'y_pred': numpy of array of predicted probabilities: shape (num_rows, num_labels)
        """
        preds_dict = {}
        for task in self.tasks:
            current_size = len(self.val_dataloader.loader_dict[task].dataset)
            preds_dict[task] = {
                'y_true': np.zeros([current_size]),
                'y_pred': np.zeros([current_size, self.task_dict[task]])
            }

        self.model = self.model.to(device)
        self.model.eval()

        with torch.no_grad():
            index_dict = {task: 0 for task in self.tasks}
            for step, batch in enumerate(progress_bar(self.val_dataloader, leave=False)):
                task_type, (x, y) = batch
                x = self._return_input_on_device(x, device)
                y = y.type(torch.LongTensor)
                y = y.to(device)

                output = self.model(x)

                y_pred = F.softmax(output[task_type], dim=1).cpu().numpy()

                y_true = y.cpu().numpy()

                current_index = index_dict[task_type]

                num_rows = self._get_num_rows(x)

                preds_dict[task_type]['y_true'][current_index: current_index + num_rows] = y_true
                preds_dict[task_type]['y_pred'][current_index: current_index + num_rows, :] = y_pred

                index_dict[task_type] += num_rows

        return preds_dict

    def _return_input_on_device(self, x, device):
        return x.to(device)

    def _get_num_rows(self, x):
        return x.size(0)


class MultiInputMultiTaskLearner(MultiTaskLearner):
    """
    Multi Input subclass of MultiTaskLearner class

    Parameters
    ----------
    model: torch.nn.Module
        PyTorch model to use with the Learner
    train_dataloader: MultiDatasetLoader
        dataloader for all of the training data
    val_dataloader: MultiDatasetLoader
        dataloader for all of the validation data
    task_dict: dict
        dictionary with all of the tasks as keys and the number of unique labels as the values

    Notes
    -----
    Multi-input datasets should return x's as a tuple/list so that each element
    can be sent to the appropriate device before being sent to the model
    see tonks.vision.dataset's TonksImageDataset class for an example
    """

    def _return_input_on_device(self, x, device):
        """
        Send all model inputs to the appropriate device (GPU or CPU)
        when the inputs are in a dictionary format.

        Parameters
        ----------
        x: dict
            Output of a dataloader where the dataset generator groups multiple
            model inputs (such as multiple images) into a dictionary. Example
            `{'full_img':some_tensor,'crop_img':some_tensor}`

        """
        for k, v in x.items():
            x[k] = v.to(device)
        return x

    def _get_num_rows(self, x):
        return x[next(iter(x))].size(0)



import numpy as np
from PIL import Image
import torch
from torch.utils.data import Dataset

from tonks.vision.config import cropped_transforms, full_img_transforms
from tonks.vision.helpers import center_crop_pil_image


class TonksEnsembleDataset(Dataset):
    """
    Load image and text data specifically for an ensemble model

    Parameters
    ----------
    text_inputs: pandas Series
        the text to be used
    img_inputs: pandas Series
        the paths to images to be used
    y: list
        A list of lists of dummy-encoded categories
    tokenizer: pretrained BERT Tokenizer
        BERT tokenizer likely from `transformers`
    max_seq_length: int (defaults to 128)
        Maximum number of tokens to allow
    transform: str or list of PyTorch transforms
        specifies how to preprocess the full image for a Tonks image model
        To use the built-in Tonks image transforms, use the strings: `train` or `val`
        To use custom transformations supply a list of PyTorch transforms.
    crop_transform: str or list of PyTorch transforms
        specifies how to preprocess the center cropped image for a Tonks image model
        To use the built-in Tonks image transforms, use strings `train` or `val`
        To use custom transformations supply a list of PyTorch transforms.
    """
    def __init__(self,
                 text_inputs,
                 img_inputs,
                 y,
                 tokenizer,
                 max_seq_length=128,
                 transform='train',
                 crop_transform='train'):
        self.text_inputs = text_inputs
        self.img_inputs = img_inputs
        self.y = y
        self.tokenizer = tokenizer
        self.max_seq_length = max_seq_length

        if transform == 'train' or 'val':
            self.transform = full_img_transforms[transform]
        else:
            self.transform = transform

        if crop_transform == 'train' or 'val':
            self.crop_transform = cropped_transforms[crop_transform]
        else:
            self.crop_transform = crop_transform

    def __getitem__(self, index):
        """Return dict of PyTorch tensors for preprocessed images and text and tensor of labels"""
        # Text processing
        x_text = self.text_inputs[index].replace('\n', ' ').replace('\r', ' ')

        tokenized_x = (
            ['[CLS]']
            + self.tokenizer.tokenize(x_text)[:self.max_seq_length - 2]
            + ['[SEP]']
        )

        input_ids = self.tokenizer.convert_tokens_to_ids(tokenized_x)

        padding = [0] * (self.max_seq_length - len(input_ids))
        input_ids += padding
        assert len(input_ids) == self.max_seq_length

        bert_text = torch.from_numpy(np.array(input_ids))

        # Image processing
        full_img = Image.open(self.img_inputs[index]).convert('RGB')

        cropped_img = center_crop_pil_image(full_img)

        full_img = self.transform(full_img)
        cropped_img = self.crop_transform(cropped_img)

        y_output = torch.from_numpy(np.array(self.y[index]))

        return {'bert_text': bert_text,
                'full_img': full_img,
                'crop_img': cropped_img}, y_output

    def __len__(self):
        return len(self.text_inputs)



import numpy as np
import torch
from torch.utils.data import Dataset


class TonksTextDataset(Dataset):
    """
    Load data for use with a BERT model

    Parameters
    ----------
    x: pandas Series
        the text to be used
    y: list
        A list of lists of dummy-encoded categories
    tokenizer: pretrained BERT Tokenizer
        BERT tokenizer likely from `transformers`
    max_seq_length: int (defaults to 128)
        Maximum number of tokens to allow
    """
    def __init__(self, x, y, tokenizer, max_seq_length=128):
        self.x = x
        self.y = y
        self.tokenizer = tokenizer
        self.max_seq_length = max_seq_length

    def __getitem__(self, index):
        """Return text as PyTorch tensor of token ids and tensor of labels"""
        text_x = self.x[index].replace('\n', ' ').replace('\r', ' ')

        tokenized_x = (
            ['[CLS]']
            + self.tokenizer.tokenize(text_x)[:self.max_seq_length - 2]
            + ['[SEP]']
        )

        input_ids = self.tokenizer.convert_tokens_to_ids(tokenized_x)

        padding = [0] * (self.max_seq_length - len(input_ids))
        input_ids += padding
        assert len(input_ids) == self.max_seq_length

        x_input = torch.from_numpy(np.array(input_ids))

        y_output = torch.from_numpy(np.array(self.y[index]))

        return x_input, y_output

    def __len__(self):
        return len(self.x)



from creevey.ops.image import centercrop
import numpy as np
from PIL import Image
import torch.nn as nn


def center_crop_pil_image(img):
    """
    Helper function to crop the center out of images.

    Utilizes the centercrop function from `creevey`

    Parameters
    ----------
    img: array
        PIL image array

    Returns
    -------
    PIL.Image: Slice of input image corresponding to a cropped area around the center
    """
    img = np.array(img)
    cropped_img = centercrop(img, reduction_factor=.4)
    return Image.fromarray(cropped_img)


class _Identity(nn.Module):
    """
    Used to pass penultimate layer features to the the ensemble

    Motivation for this is that the features from the penultimate layer
    are likely more informative than the 1000 way softmax that was used
    in the multi_output_model_v2.
    """
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x


def _dense_block(in_f, out_f, reg):
    return nn.Sequential(nn.Linear(in_f, out_f),
                         nn.BatchNorm1d(out_f, eps=reg),
                         nn.ReLU()
                         )



import numpy as np
from PIL import Image
import torch
from torch.utils.data import Dataset

from tonks.vision.config import cropped_transforms, full_img_transforms
from tonks.vision.helpers import center_crop_pil_image


class TonksImageDataset(Dataset):
    """
    Load data specifically for use with a image models

    Parameters
    ----------
    x: pandas Series
        file paths to stored images
    y: list
        A list of dummy-encoded categories
        For instance, y might be [0,1,2,0] for a 3 class problem with 4 samples
    transform: str or list of PyTorch transforms
        specifies how to preprocess the full image for a Tonks image model
        To use the built-in Tonks image transforms, use the strings: `train` or `val`
        To use custom transformations supply a list of PyTorch transforms
    crop_transform: str or list of PyTorch transforms
        specifies how to preprocess the center cropped image for a Tonks image model
        To use the built-in Tonks image transforms, use strings `train` or `val`
        To use custom transformations supply a list of PyTorch transforms
    """
    def __init__(self,
                 x,
                 y,
                 transform='train',
                 crop_transform='train'):
        self.x = x
        self.y = y

        if transform == 'train' or 'val':
            self.transform = full_img_transforms[transform]
        else:
            self.transform = transform

        if crop_transform == 'train' or 'val':
            self.crop_transform = cropped_transforms[crop_transform]
        else:
            self.crop_transform = crop_transform

    def __getitem__(self, index):
        """Return tuple of images as PyTorch tensors and and tensor of labels"""
        label = self.y[index]
        full_img = Image.open(self.x[index]).convert('RGB')

        cropped_img = center_crop_pil_image(full_img)

        full_img = self.transform(full_img)
        cropped_img = self.crop_transform(cropped_img)

        label = torch.from_numpy(np.array(label)).float()

        return {'full_img': full_img,
                'crop_img': cropped_img}, label

    def __len__(self):
        return len(self.x)



from setuptools import find_packages, setup

with open('tonks/_version.py') as version_file:
    exec(version_file.read())

with open('README.md') as r:
    readme = r.read()

with open('LICENSE.txt') as l:
    license = l.read()

setup(
    name='tonks',
    version=__version__,
    description='General purpose multi-task classification library',
    long_description=readme+'\n\n\nLicense\n-------\n'+license,
    long_description_content_type='text/markdown',
    author='Nicole Carlson, Michael Sugimura',
    url='https://github.com/shoprunner/tonks',
    license='BSD-3-Clause',
    data_files=[('', ['LICENSE.txt'])],
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
        'creevey',
        'fastprogress',
        'joblib',
        'numpy',
        'Pillow<7.0.0',
        'transformers>=2.3.0',
        'scikit-learn',
        'torch==1.2.0',
        'torchvision==0.2.1',
    ],
    extra_requires={
        'dev': [
            'flake8',
            'flake8-docstrings',
            'flake8-import-order',
            'm2r',
            'pydocstyle<4.0.0',
            'pytest',
            'pytest-cov',
            'sphinx-rtd-theme==0.4.3'
        ]
    },
)



import copy
from pathlib import Path

import torch
import torch.nn as nn
from transformers.modeling_bert import BertModel, BertPreTrainedModel


class BertForMultiTaskClassification(BertPreTrainedModel):
    """
    PyTorch BERT class for multitask learning. This model allows you to load
    in some pretrained tasks in addition to creating new ones.

    Examples
    --------
    To instantiate a completely new instance of BertForMultiTaskClassification
    and load the weights into this architecture you can use the `from_pretrained`
    method of the base class by specifying the name of the weights to load, e.g.::

        model = BertForMultiTaskClassification.from_pretrained(
            'bert-base-uncased',
            new_task_dict=new_task_dict
        )

        # DO SOME TRAINING

        model.save(SOME_FOLDER, SOME_MODEL_ID)

    To instantiate an instance of BertForMultiTaskClassification that has layers for
    pretrained tasks and new tasks, you would do the following::

        model = BertForMultiTaskClassification.from_pretrained(
            'bert-base-uncased',
            pretrained_task_dict=pretrained_task_dict,
            new_task_dict=new_task_dict
        )

        model.load(SOME_FOLDER, SOME_MODEL_DICT)

        # DO SOME TRAINING

    Parameters
    ----------
    config: json file
        Defines the BERT model architecture.
        Note: you will most likely be instantiating the class with the `from_pretrained` method
        so you don't need to come up with your own config.
    pretrained_task_dict: dict
        dictionary mapping each pretrained task to the number of labels it has
    new_task_dict: dict
        dictionary mapping each new task to the number of labels it has
    dropout: float
        dropout percentage for Dropout layer
    """

    def __init__(self, config, pretrained_task_dict=None, new_task_dict=None, dropout=1e-1):
        super(BertForMultiTaskClassification, self).__init__(config)
        self.bert = BertModel(config)

        self.dropout = torch.nn.Dropout(dropout)

        if pretrained_task_dict is not None:
            pretrained_layers = {}
            for key, task_size in pretrained_task_dict.items():
                pretrained_layers[key] = nn.Linear(config.hidden_size, task_size)
            self.pretrained_classifiers = nn.ModuleDict(pretrained_layers)
        if new_task_dict is not None:
            new_layers = {}
            for key, task_size in new_task_dict.items():
                new_layers[key] = nn.Linear(config.hidden_size, task_size)
            self.new_classifiers = nn.ModuleDict(new_layers)

    def forward(self, tokenized_input):
        """
        Defines forward pass for Bert model

        Parameters
        ----------
        tokenized_input: torch tensor of integers
            integers represent tokens for each word

        Returns
        ----------
        A dictionary mapping each task to its logits
        """
        outputs = self.bert(tokenized_input)

        pooled_output = self.dropout(outputs[1])

        logit_dict = {}
        if hasattr(self, 'pretrained_classifiers'):
            for key, classifier in self.pretrained_classifiers.items():
                logit_dict[key] = classifier(pooled_output)
        if hasattr(self, 'new_classifiers'):
            for key, classifier in self.new_classifiers.items():
                logit_dict[key] = classifier(pooled_output)

        return logit_dict

    def freeze_bert(self):
        """Freeze all core Bert layers"""
        for param in self.bert.parameters():
            param.requires_grad = False

    def freeze_pretrained_classifiers_and_bert(self):
        """Freeze pretrained classifier layers and core Bert layers"""
        self.freeze_bert()
        if hasattr(self, 'pretrained_classifiers'):
            for param in self.pretrained_classifiers.parameters():
                param.requires_grad = False
        else:
            print('There are no pretrained_classifier layers to be frozen.')

    def unfreeze_pretrained_classifiers(self):
        """Unfreeze pretrained classifier layers"""
        if hasattr(self, 'pretrained_classifiers'):
            for param in self.pretrained_classifiers.parameters():
                param.requires_grad = True
        else:
            print('There are no pretrained_classifier layers to be unfrozen.')

    def unfreeze_pretrained_classifiers_and_bert(self):
        """Unfreeze pretrained classifiers and core Bert layers"""
        for param in self.bert.parameters():
            param.requires_grad = True

        self.unfreeze_pretrained_classifiers()

    def save(self, folder, model_id):
        """
        Saves the model state dicts to a specific folder.
        Each part of the model is saved separately to allow for
        new classifiers to be added later.

        Note: if the model has `pretrained_classifiers` and `new_classifers`,
        they will be combined into the `pretrained_classifiers_dict`.

        Parameters
        ----------
        folder: str or Path
            place to store state dictionaries
        model_id: int
            unique id for this model

        Side Effects
        ------------
        saves three files:
            - folder / f'bert_dict_{model_id}.pth'
            - folder / f'dropout_dict_{model_id}.pth'
            - folder / f'pretrained_classifiers_dict_{model_id}.pth'
        """
        if not hasattr(self, 'pretrained_classifiers'):
            classifiers_to_save = copy.deepcopy(self.new_classifiers)
        else:
            # PyTorch's update method isn't working because it doesn't think ModuleDict is a Mapping
            classifiers_to_save = copy.deepcopy(self.pretrained_classifiers)
            for key, module in self.new_classifiers.items():
                classifiers_to_save[key] = module

        folder = Path(folder)
        folder.mkdir(parents=True, exist_ok=True)

        torch.save(
            self.bert.state_dict(),
            folder / f'bert_dict_{model_id}.pth'
        )
        torch.save(
            self.dropout.state_dict(),
            folder / f'dropout_dict_{model_id}.pth'
        )

        torch.save(
            classifiers_to_save.state_dict(),
            folder / f'pretrained_classifiers_dict_{model_id}.pth'
        )

    def load(self, folder, model_id):
        """
        Loads the model state dicts from a specific folder.

        Parameters
        ----------
        folder: str or Path
            place where state dictionaries are stored
        model_id: int
            unique id for this model

        Side Effects
        ------------
        loads from three files:
            - folder / f'bert_dict_{model_id}.pth'
            - folder / f'dropout_dict_{model_id}.pth'
            - folder / f'pretrained_classifiers_dict_{model_id}.pth'
        """
        folder = Path(folder)

        if torch.cuda.is_available():
            self.bert.load_state_dict(torch.load(folder / f'bert_dict_{model_id}.pth'))
            self.dropout.load_state_dict(torch.load(folder / f'dropout_dict_{model_id}.pth'))
            self.pretrained_classifiers.load_state_dict(
                torch.load(folder / f'pretrained_classifiers_dict_{model_id}.pth')
            )
        else:
            self.bert.load_state_dict(
                torch.load(
                    folder / f'bert_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )
            self.dropout.load_state_dict(
                torch.load(
                    folder / f'dropout_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )
            self.pretrained_classifiers.load_state_dict(
                torch.load(
                    folder / f'pretrained_classifiers_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )

    def export(self, folder, model_id, model_name=None):
        """
        Exports the entire model state dict to a specific folder.

        Note: if the model has `pretrained_classifiers` and `new_classifers`,
        they will be combined into the `pretrained_classifiers` attribute before being saved.

        Parameters
        ----------
        folder: str or Path
            place to store state dictionaries
        model_id: int
            unique id for this model
        model_name: str (defaults to None)
            Name to store model under, if None, will default to `multi_task_bert_{model_id}.pth`

        Side Effects
        ------------
        saves one file:
            - folder / model_name
        """
        hold_new_classifiers = copy.deepcopy(self.new_classifiers)
        hold_pretrained_classifiers = None
        if not hasattr(self, 'pretrained_classifiers'):
            self.pretrained_classifiers = copy.deepcopy(self.new_classifiers)
        else:
            hold_pretrained_classifiers = copy.deepcopy(self.pretrained_classifiers)
            # PyTorch's update method isn't working because it doesn't think ModuleDict is a Mapping
            for key, module in self.new_classifiers.items():
                self.pretrained_classifiers[key] = module

        del self.new_classifiers

        if model_name is None:
            model_name = f'multi_task_bert_{model_id}.pth'

        folder = Path(folder)
        folder.mkdir(parents=True, exist_ok=True)

        torch.save(
            self.state_dict(),
            folder / model_name
        )
        if hold_pretrained_classifiers is not None:
            self.pretrained_classifiers = hold_pretrained_classifiers
        else:
            del self.pretrained_classifiers
        self.new_classifiers = hold_new_classifiers



from pathlib import Path

import joblib
import torch
import torch.nn as nn
from torchvision import models as torch_models
from transformers.modeling_bert import BertConfig, BertModel

from tonks.exc import TonksError
from tonks.vision.helpers import _dense_block, _Identity


class BertResnetEnsembleForMultiTaskClassification(nn.Module):
    """
    PyTorch ensemble class for multitask learning consisting of a text and image models

    This model is made up of multiple component models:
    - for text: Google's BERT model
    - for images: multiple ResNet50's (the exact number depends on how
    the image model tasks were split up)

    You may need to train the component image and text models first
    before combining them into an ensemble model to get good results.

    Note: For explicitness, `vanilla` refers to the
    `transformers` BERT or `PyTorch` ResNet50 weights while
    `pretrained` refers to previously trained Tonks weights.

    Examples
    --------
    The ensemble model should be used with pretrained
    BERT and ResNet50 component models.
    To initialize a model in this way::

        image_task_dict = {
            'color_pattern': {
                'color': color_train_df['labels'].nunique(),
                'pattern': pattern_train_df['labels'].nunique()
            },
            'dress_sleeve': {
                'dress_length': dl_train_df['labels'].nunique(),
                'sleeve_length': sl_train_df['labels'].nunique()
            },
            'season': {
                'season': season_train_df['labels'].nunique()
            }
        }
        model = BertResnetEnsembleForMultiTaskClassification(
            image_task_dict=image_task_dict
        )

        resnet_model_id_dict = {
            'color_pattern': 'SOME_RESNET_MODEL_ID1',
            'dress_sleeve': 'SOME_RESNET_MODEL_ID2',
            'season': 'SOME_RESNET_MODEL_ID3'
        }

        model.load_core_models(
            folder='SOME_FOLDER',
            bert_model_id='SOME_BERT_MODEL_ID',
            resnet_model_id_dict=resnet_model_id_dict
        )

        # DO SOME TRAINING

        model.save(SOME_FOLDER, SOME_MODEL_ID)

        # OR

        model.export(SOME_FOLDER, SOME_MODEL_ID)

    Parameters
    ----------
    image_task_dict: dict
        dictionary mapping each pretrained ResNet50 models to a dictionary
        of the tasks it was trained on
    dropout: float
        dropout percentage for Dropout layer
    """

    def __init__(
        self,
        image_task_dict=None,
        dropout=1e-1
    ):
        super(BertResnetEnsembleForMultiTaskClassification, self).__init__()

        # Define text architecture
        config = BertConfig()
        self.bert = BertModel(config)
        self.dropout = torch.nn.Dropout(dropout)

        self.image_task_dict = image_task_dict
        self.text_task_dict = self.create_text_dict(image_task_dict)

        # Define image architecture
        image_resnets = {}
        image_dense_layers = {}
        ensemble_layers = {}
        for key in self.image_task_dict.keys():
            resnet = torch_models.resnet50(pretrained=False)
            resnet.fc = _Identity()
            image_resnets[key] = resnet
            image_dense_layers[key] = nn.Sequential(
                _dense_block(2048*2, 1024, 2e-3),
                _dense_block(1024, 512, 2e-3),
                _dense_block(512, 256, 2e-3)
            )

            # Define final ensemble before classifier layers
            # The input is size 768 from BERT and 256 from ResNet50 models
            # so the total size is 1024
            ensemble_layers[key] = nn.Sequential(
                _dense_block(1024, 512, 2e-3),
                _dense_block(512, 512, 2e-3),
                _dense_block(512, 256, 2e-3),
            )

        self.image_resnets = nn.ModuleDict(image_resnets)
        self.image_dense_layers = nn.ModuleDict(image_dense_layers)
        self.ensemble_layers = nn.ModuleDict(ensemble_layers)

        pretrained_layers = {}
        for key, task_size in self.text_task_dict.items():
            pretrained_layers[key] = nn.Linear(256, task_size)
        self.classifiers = nn.ModuleDict(pretrained_layers)

    def forward(self, x):
        """
        Defines forward pass for ensemble model

        Parameters
        ----------
        x: dict
            dictionary of torch tensors with keys:
                - `bert_text`: integers mapping to BERT vocabulary
                - `full_img`: tensor of full image
                - `crop_img`: tensor of cropped image

        Returns
        ----------
        A dictionary mapping each task to its logits
        """
        bert_output = self.bert(x['bert_text'])

        pooled_output = self.dropout(bert_output[1])

        logit_dict = {}

        for key in self.image_task_dict.keys():
            full_img = self.image_resnets[key](x['full_img'])
            crop_img = self.image_resnets[key](x['crop_img'])
            full_crop_combined = torch.cat((full_img, crop_img), 1)
            dense_layer_output = self.image_dense_layers[key](full_crop_combined)
            ensemble_input = torch.cat((pooled_output, dense_layer_output), 1)
            ensemble_layer_output = self.ensemble_layers[key](ensemble_input)

            for task in self.image_task_dict[key].keys():
                classifier = self.classifiers[task]
                logit_dict[task] = classifier(ensemble_layer_output)

        return logit_dict

    def freeze_bert(self):
        """Freeze all core BERT layers"""
        for param in self.bert.parameters():
            param.requires_grad = False

    def freeze_resnets(self):
        """Freeze all core ResNet models layers"""
        for key in self.image_resnets.keys():
            for param in self.image_resnets[key].parameters():
                param.requires_grad = False
            for param in self.image_dense_layers[key].parameters():
                param.requires_grad = False

    def freeze_ensemble_layers(self):
        """Freeze all final ensemble layers"""
        for key in self.ensemble_layers.keys():
            for param in self.ensemble_layers[key].parameters():
                param.requires_grad = False

    def freeze_classifiers_and_core(self):
        """Freeze pretrained classifier layers and core BERT/ResNet layers"""
        self.freeze_bert()
        self.freeze_resnets()
        self.freeze_ensemble_layers()
        for param in self.classifiers.parameters():
            param.requires_grad = False

    def unfreeze_classifiers(self):
        """Unfreeze pretrained classifier layers"""
        for param in self.classifiers.parameters():
            param.requires_grad = True

    def unfreeze_classifiers_and_core(self):
        """Unfreeze pretrained classifiers and core BERT/ResNet layers"""
        for param in self.bert.parameters():
            param.requires_grad = True
        for key in self.image_resnets.keys():
            for param in self.image_resnets[key].parameters():
                param.requires_grad = True
            for param in self.image_dense_layers[key].parameters():
                param.requires_grad = True
            for param in self.ensemble_layers[key].parameters():
                param.requires_grad = True

        self.unfreeze_classifiers()

    def save(self, folder, model_id):
        """
        Saves the model state dicts to a specific folder.
        Each part of the model is saved separately,
        along with the image_task_dict, which is needed to reinstantiate the model.

        Parameters
        ----------
        folder: str or Path
            place to store state dictionaries
        model_id: int
            unique id for this model

        Side Effects
        ------------
        saves six files:
            - folder / f'bert_dict_{model_id}.pth'
            - folder / f'dropout_dict_{model_id}.pth'
            - folder / f'image_resnets_dict_{model_id}.pth'
            - folder / f'image_dense_layers_dict_{model_id}.pth'
            - folder / f'ensemble_layers_dict_{model_id}.pth'
            - folder / f'classifiers_dict_{model_id}.pth'
        """
        folder = Path(folder)
        folder.mkdir(parents=True, exist_ok=True)

        # BERT model
        torch.save(
            self.bert.state_dict(),
            folder / f'bert_dict_{model_id}.pth'
        )
        torch.save(
            self.dropout.state_dict(),
            folder / f'dropout_dict_{model_id}.pth'
        )

        # ResNet model(s)
        torch.save(
            self.image_resnets.state_dict(),
            folder / f'image_resnets_dict_{model_id}.pth'
        )
        torch.save(
            self.image_dense_layers.state_dict(),
            folder / f'image_dense_layers_dict_{model_id}.pth'
        )

        # Ensemble layers
        torch.save(
            self.ensemble_layers.state_dict(),
            folder / f'ensemble_layers_dict_{model_id}.pth'
        )

        # Classifier layers
        torch.save(
            self.classifiers.state_dict(),
            folder / f'classifiers_dict_{model_id}.pth'
        )

        # image_task_dict
        joblib.dump(self.image_task_dict, folder / f'image_task_dict_{model_id}.pickle')

    def load(self, folder, model_id):
        """
        Loads the model state dicts for ensemble model
        from a specific folder. This will load all the model
        components including the final ensemble and existing
        pretrained `classifiers`.

        Parameters
        ----------
        folder: str or Path
            place where state dictionaries are stored
        model_id: int
            unique id for this model

        Side Effects
        ------------
        loads from six files:
            - folder / f'bert_dict_{model_id}.pth'
            - folder / f'dropout_dict_{model_id}.pth'
            - folder / f'image_resnets_dict_{model_id}.pth'
            - folder / f'dense_layers_dict_{model_id}.pth'
            - folder / f'ensemble_layers_dict_{model_id}.pth'
            - folder / f'classifiers_dict_{model_id}.pth'

        """
        folder = Path(folder)

        if torch.cuda.is_available():
            self.bert.load_state_dict(
                torch.load(folder / f'bert_dict_{model_id}.pth'))
            self.dropout.load_state_dict(
                torch.load(folder / f'dropout_dict_{model_id}.pth'))

            self.image_resnets.load_state_dict(
                torch.load(folder / f'image_resnets_dict_{model_id}.pth'))
            self.dense_layers.load_state_dict(
                torch.load(folder / f'image_dense_layers_dict_{model_id}.pth'))

            self.ensemble_layers.load_state_dict(
                torch.load(folder / f'ensemble_layers_dict_{model_id}.pth'))
            self.classifiers.load_state_dict(
                torch.load(folder / f'classifiers_dict_{model_id}.pth')
            )
        else:
            self.bert.load_state_dict(
                torch.load(
                    folder / f'bert_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )
            self.dropout.load_state_dict(
                torch.load(
                    folder / f'dropout_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )

            self.resnet.load_state_dict(
                torch.load(
                    folder / f'image_resnets_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )
            self.dense_layers.load_state_dict(
                torch.load(
                    folder / f'image_dense_layers_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )

            self.final_ensemble.load_state_dict(
                torch.load(
                    folder / f'ensemble_layers_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )

            self.classifiers.load_state_dict(
                torch.load(
                    folder / f'classifiers_dict_{model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )

    def load_core_models(self, folder, bert_model_id, resnet_model_id_dict):
        """
        Loads the weights from pretrained BERT and ResNet50 Tonks models

        Does not load weights from the final ensemble and classifier layers.
        use case is for loading SR_pretrained component BERT and image model
        weights into a new ensemble model.

        Parameters
        ----------
        folder: str or Path
            place where state dictionaries are stored
        bert_model_id: int
            unique id for pretrained BERT text model
        resnet_model_id: int
            unique id for pretrained image model

        Side Effects
        ------------
        loads from four files:
            - folder / f'bert_dict_{bert_model_id}.pth'
            - folder / f'dropout_dict_{bert_model_id}.pth'
            - folder / f'resnet_dict_{resnet_model_id}.pth'
            - folder / f'dense_layers_dict_{resnet_model_id}.pth'
        """
        folder = Path(folder)

        if torch.cuda.is_available():
            self.bert.load_state_dict(
                torch.load(folder / f'bert_dict_{bert_model_id}.pth'))
            self.dropout.load_state_dict(
                torch.load(folder / f'dropout_dict_{bert_model_id}.pth'))

            for key, model_id in resnet_model_id_dict.items():
                self.image_resnets[key].load_state_dict(
                    torch.load(folder / f'resnet_dict_{model_id}.pth')
                )
                self.image_dense_layers[key].load_state_dict(
                    torch.load(folder / f'dense_layers_dict_{model_id}.pth')
                )

        else:
            self.bert.load_state_dict(
                torch.load(
                    folder / f'bert_dict_{bert_model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )
            self.dropout.load_state_dict(
                torch.load(
                    folder / f'dropout_dict_{bert_model_id}.pth',
                    map_location=lambda storage,
                    loc: storage
                )
            )

            for key, model_id in resnet_model_id_dict.items():
                self.image_resnets[key].load_state_dict(
                    torch.load(folder / f'resnet_dict_{model_id}.pth'),
                    map_location=lambda storage,
                    loc: storage
                )
                self.image_dense_layers[key].load_state_dict(
                    torch.load(folder / f'dense_layers_dict_{model_id}.pth'),
                    map_location=lambda storage,
                    loc: storage
                )

    def export(self, folder, model_id, model_name=None):
        """
        Exports the entire model state dict to a specific folder,
        along with the image_task_dict, which is needed to reinstantiate the model.

        Parameters
        ----------
        folder: str or Path
            place to store state dictionaries
        model_id: int
            unique id for this model
        model_name: str (defaults to None)
            Name to store model under, if None, will default to `multi_task_ensemble_{model_id}.pth`

        Side Effects
        ------------
        saves two files:
            - folder / f'multi_task_ensemble_{model_id}.pth'
            - folder / f'image_task_dict_{model_id}.pickle'
        """
        folder = Path(folder)
        folder.mkdir(parents=True, exist_ok=True)

        if model_name is None:
            model_name = f'multi_task_ensemble_{model_id}.pth'

        torch.save(
            self.state_dict(),
            folder / model_name
        )

        joblib.dump(self.image_task_dict, folder / f'image_task_dict_{model_id}.pickle')

    @staticmethod
    def create_text_dict(image_task_dict):
        """Create a task dict for the text model from the image task dict"""
        text_task_dict = {}
        for joint_task in image_task_dict.keys():
            for task, task_size in image_task_dict[joint_task].items():
                if task in text_task_dict.keys():
                    raise TonksError(
                        'Task {} is in multiple models. Each task can only be in one image model.'
                        .format(task)
                    )
                text_task_dict[task] = task_size

        return text_task_dict



import pytest
import torch


@pytest.fixture()
def test_label_dictionary():
    color_list = ['beige', 'black']
    dress_length_list = ['mini', 'maxi']
    pattern_list = ['dotted', 'floral']
    sleeve_length_list = ['sleeveless', 'long_sleeve']
    attribute_label_dict = {
        'color': color_list,
        'dress_length': dress_length_list,
        'pattern': pattern_list,
        'sleeve_length': sleeve_length_list
    }

    return attribute_label_dict


@pytest.fixture()
def test_raw_logits():
    return {
        'color': torch.tensor([[0, 5.0]]),
        'dress_length': torch.tensor([[0, 5.0]]),
        'pattern': torch.tensor([[5.0, 0]]),
        'sleeve_length': torch.tensor([[5.0, 0]])
    }



from ._version import __version__

from tonks.dataloader import MultiDatasetLoader
from tonks.ensemble import *
from tonks.learner import MultiTaskLearner, MultiInputMultiTaskLearner
from tonks.text import *
from tonks.vision import *
